#include <Rotary.h>
#include <SPI.h>
#include <Adafruit_GFX.h>
#include <Adafruit_ST7735.h>

void testFunction(int_fast32_t, int_fast32_t, int_fast32_t, void (*)(int_fast32_t));
void f(int_fast32_t);

const int W_CLK = 10;
const int FQ_UD = 9;
const int DATA = 7;
const int RESET = 12;

// TFT
#define cs   5
#define rst  4
#define dc   6

Adafruit_ST7735 tft = Adafruit_ST7735(cs, dc, rst);

Rotary r = Rotary(2, 3); // sets the pins the rotary encoder uses.

const int buttonPin = 8;

int buttonState = 0;
int_fast32_t rx = 10000000; // Starting frequency of VFO 10.000 MHz
int_fast32_t rx_old;        // variable to hold the updated frequency
int_fast32_t rx_ofs;        // command frequency for DDS incl. possible IF offset
int_fast32_t step1 = 1000;   // starting VFO update increment in HZ.

float rx_disp;
int  stp = 2;                           // Начальный шаг 100 Hz
int  stp_old;

////////////////////////////////////////////////////////////
void setup() {
  Serial.begin(9600);
  tft.initR(INITR_BLACKTAB);
  tft.setRotation(tft.getRotation() - 1); ///// Крутим дисплей на  180 Градусов
  tft.setTextSize(1);    //1-8
  tft.fillScreen(ST7735_BLACK);
  tft.setCursor(20, 2);
  tft.setTextColor(ST7735_GREEN);
  tft.cp437(true);
  tft.print(utf8rus("DDS Sweep RA4FHE"));
  tft.setTextSize(2);
  tft.setCursor(15, 25);
  tft.setTextColor(ST7735_BLUE);
  tft.print(utf8rus("Частота:"));

  PCICR |= (1 << PCIE2);
  PCMSK2 |= (1 << PCINT18) | (1 << PCINT19);
  sei();
  pinMode(buttonPin, INPUT);
  pinMode(FQ_UD, OUTPUT);
  pinMode(W_CLK, OUTPUT);
  pinMode(DATA, OUTPUT);
  pinMode(RESET, OUTPUT);
  digitalWrite(RESET, HIGH);
  digitalWrite(RESET, LOW);
  digitalWrite(8, HIGH);

}

//////////////////// Основной цикл ////////////////////////////////////////
void loop() {
  read_Step();
  read_Step_Select();
  buttonState = digitalRead(buttonPin);
  rx_disp = rx;
  rx_disp = rx_disp / 1000000;
  tft.fillRect(5, 45, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
  tft.setCursor(15, 50);
  tft.setTextColor(ST7735_WHITE);
  tft.print(rx_disp, 5 );
  tft.print(" MHz");

  rx_old = rx;
  rx_ofs = rx;
  const int ciStep = 15000; /////////////// Девиация 30 KHz
  testFunction(rx_ofs - ciStep, rx_ofs + ciStep, 1000, &f); //////////// Шаг Качания 

}

////////////////////////////////////////////////////////////
// frequency calc
void sendFrequency(double frequency) {
  int32_t freq = frequency * 4294967295 / 125000000; // note 125 MHz clock on AD9850.  You can make 'slight' tuning variations here by adjusting the clock frequency.
  for (int b = 0; b < 4; b++, freq >>= 8) {
    tfr_byte(freq & 0xFF);
  }
  tfr_byte(0);   // Final control byte, all 0 for AD9850 chip
  digitalWrite(FQ_UD, HIGH);
  digitalWrite(FQ_UD, LOW); // Done!  Should see output
}

////////////////////////////////////////////////////////////
// transfers a byte, a bit at a time, LSB first to the AD9850 via serial DATA line
void tfr_byte(byte data) {
  for (int i = 0; i < 8; i++, data >>= 1)
  {
    digitalWrite(DATA, data & 0x01);
    digitalWrite(W_CLK, HIGH);
    digitalWrite(W_CLK, LOW);  //after each bit sent, CLK is pulsed high
  }
}

/////////////////////////////////////////////////////////////////
void read_Step_Select() {
  if (stp == 7)
  {
    stp = 1;
  }
  if (stp == 0)
  {
    stp = 6;
  }

  if (stp != stp_old)
  {
    if (stp == 1)
    {
      step1 = 10;
      tft.fillRect(5, 100, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
      tft.setCursor(15, 80);
      tft.setTextColor(ST7735_CYAN);
      tft.print(utf8rus("Шаг:"));
      tft.setCursor(15, 105);
      tft.setTextColor(ST7735_WHITE);
      tft.print("10 Hz");

    }
    else if (stp == 2)
    {
      step1 = 100;
      tft.fillRect(5, 100, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
      tft.setCursor(15, 80);
      tft.setTextColor(ST7735_CYAN);
      tft.print(utf8rus("Шаг:"));
      tft.setCursor(15, 105);
      tft.setTextColor(ST7735_WHITE);
      tft.print("100 Hz");
    }
    else if (stp == 3)
    {
      step1 = 1000;
      tft.fillRect(5, 100, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
      tft.setCursor(15, 80);
      tft.setTextColor(ST7735_CYAN);
      tft.print(utf8rus("Шаг:"));
      tft.setCursor(15, 105);
      tft.setTextColor(ST7735_WHITE);
      tft.print("1 KHz");
    }
    else if (stp == 4)
    {
      step1 = 10000;
      tft.fillRect(5, 100, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
      tft.setCursor(15, 80);
      tft.setTextColor(ST7735_CYAN);
      tft.print(utf8rus("Шаг:"));
      tft.setCursor(15, 105);
      tft.setTextColor(ST7735_WHITE);
      tft.print("10 KHz");
    }
    else if (stp == 5)
    {
      step1 = 100000;
      tft.fillRect(5, 100, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
      tft.setCursor(15, 80);
      tft.setTextColor(ST7735_CYAN);
      tft.print(utf8rus("Шаг:"));
      tft.setCursor(15, 105);
      tft.setTextColor(ST7735_WHITE);
      tft.print("100 KHz");
    }
    else if (stp == 6)
    {
      step1 = 1000000;
      tft.fillRect(5, 100, 155, 20, ST7735_BLACK); //partly screen erasing by writing box
      tft.setCursor(15, 80);
      tft.setTextColor(ST7735_CYAN);
      tft.print(utf8rus("Шаг:"));
      tft.setCursor(15, 105);
      tft.setTextColor(ST7735_WHITE);
      tft.print("1 MHz");
    }
    delay(500);
    stp_old = stp;
  }
}

int read_Step()
{
  if (buttonState == LOW) return stp = stp + 1;
}

ISR(PCINT2_vect) {
  unsigned char result = r.process();
  if (result) {
    if (result == DIR_CW) {
      rx = rx + step1;
    }
    else {
      rx = rx - step1;
    };
    if (rx >= 40000000) {
      rx = rx_old;
    }; // UPPER VFO LIMIT 40MHz
    if  (rx <= 100000) {
      rx = rx_old;
    }; // LOWER VFO LIMIT 100KHz
  }
}
//////////////////////// RUSFONT ///////////////////////////////////////////////////

#define maxString 21
char target[maxString + 1] = "";

char *utf8rus(char *source)
{
  int i, j, k;
  unsigned char n;
  char m[2] = { '0', '\0' };

  strcpy(target, ""); k = strlen(source); i = j = 0;

  while (i < k) {
    n = source[i]; i++;

    if (n >= 0xC0) {
      switch (n) {
        case 0xD0: {
            n = source[i]; i++;
            if (n == 0x81) {
              n = 0xA8;
              break;
            }
            if (n >= 0x90 && n <= 0xBF) n = n + 0x30;
            break;
          }
        case 0xD1: {
            n = source[i]; i++;
            if (n == 0x91) {
              n = 0xB8;
              break;
            }
            if (n >= 0x80 && n <= 0x8F) n = n + 0x70;
            break;
          }
      }
    }

    m[0] = n; strcat(target, m);
    j++; if (j >= maxString) break;
  }
  return target;
}

/////////////////////////Sweep DDS>>>>>>>>>>

void f(int_fast32_t i) {
  sendFrequency(i);       // frequency for controlling the DDS
  delay(10);
}

void testFunction(int_fast32_t argMin, int_fast32_t argMax, int_fast32_t argStep, void (*fPtr)(int_fast32_t)) {
  int_fast32_t start = 0;
  for (start = argMin; start < argMax; start += argStep) {
    fPtr(start);
  }
  for (start = argMax; start > argMin; start -= argStep) {
    fPtr(start);
  }
}
